import json
from datetime import datetime
from pathlib import Path
import pkg_resources, requests
from ...utils.environment import IS_PRODUCTION
from ...utils import get_package_version


CACHED_VERSION_FILE = ".abstra/cached_versions"
EXPIRE_PERIOD = 60 * 60 * 4  # 4 hours
TIMEOUT = 5


def check_latest_version(package_name="abstra"):
    if IS_PRODUCTION:
        return

    try:
        current_version = get_package_version(package_name)
        latest_version = _get_cached_latest_version(package_name)

        highest_version = (
            _get_highest_version(current_version, latest_version)
            if isinstance(current_version, str)
            else latest_version
        )

        if highest_version == latest_version:
            print(
                f"A new version of Abstra Editor is available. Latest version is {latest_version}, but you have {current_version}."
            )
            print(f"Please run 'pip install {package_name} --upgrade' to update.")
        else:
            print(f"Abstra Editor is up to date (version {current_version}).")

            if isinstance(current_version, str) and highest_version == current_version:
                _update_cached_latest_version(package_name, current_version)

    except pkg_resources.DistributionNotFound:
        print(f"{package_name} is not installed.")

    except requests.exceptions.RequestException:
        print(f"Unable to fetch version information for {package_name} from PyPI.")

    print("\n")


def _get_cached_latest_version(package_name="abstra") -> str:
    cached_file = Path(CACHED_VERSION_FILE) / f"{package_name}.json"

    if cached_file.exists():
        try:
            with open(cached_file, "r") as f:
                cached_version = json.loads(f.readline())

            created_at = cached_version["created_at"]
            version = cached_version["version"]
        except:
            created_at = None
            version = None

        if (
            created_at is not None
            and isinstance(created_at, float)
            and version is not None
            and datetime.utcnow().timestamp() - created_at < EXPIRE_PERIOD
        ):
            return version

    response = requests.get(
        f"https://pypi.org/pypi/{package_name}/json", timeout=TIMEOUT
    )
    latest_version = response.json()["info"]["version"]

    _update_cached_latest_version(package_name, latest_version)

    return latest_version


def _update_cached_latest_version(package_name, latest_version: str):
    cached_file = Path(CACHED_VERSION_FILE) / f"{package_name}.json"

    if not cached_file.parent.exists():
        cached_file.parent.mkdir(parents=True)

    with open(cached_file, "w") as f:
        json.dump(
            {
                "version": latest_version,
                "created_at": datetime.utcnow().timestamp(),
            },
            f,
        )


def _get_highest_version(first_version: str, second_version: str):
    first_version_array = first_version.split(".")
    second_version_array = second_version.split(".")

    if (len(first_version_array) != 3) or (len(second_version_array) != 3):
        return None

    for i in range(3):
        if int(first_version_array[i]) > int(second_version_array[i]):
            return first_version
        elif int(first_version_array[i]) < int(second_version_array[i]):
            return second_version

    return None



from .execution_logs import execution_logs_repository_factory
from .connectors import connectors_api_http_client_factory
from .stage_run import stage_run_repository_factory
from .execution import execution_repository_factory
from .tables import tables_api_http_client_factory
from .authn import authn_repository_factory
from .notifications import notification_repository_factory

connectors_api_http_client = connectors_api_http_client_factory()
execution_logs_repository = execution_logs_repository_factory()
tables_api_http_client = tables_api_http_client_factory()
stage_run_repository = stage_run_repository_factory()
execution_repository = execution_repository_factory()
authn_repository = authn_repository_factory()
notification_repository = notification_repository_factory()



from datetime import datetime
from pydantic.dataclasses import dataclass
from typing import Any, List, Optional, Type, Sequence, Tuple
from .workflows import make_stage_dto
from ...repositories.project.project import (
    ProjectRepository,
    WorkflowStage,
    FormStage,
)
from ...repositories.stage_run import (
    StageRunRepository,
    StageRun,
    GetStageRunByQueryFilter,
    Pagination,
    stage_run_repository_factory,
)

from ...repositories.members import members_repository_factory
from ..guards.role_guard import RoleGuardFactory
from typing import Any, List, Optional, Sequence, Tuple, Type

import flask

from ...repositories import execution_logs_repository, stage_run_repository
from ...repositories.execution_logs import (
    LogEntry,
    StdioLogEntry,
    UnhandledExceptionLogEntry,
)
from ...repositories.project.project import FormStage, ProjectRepository, WorkflowStage
from ...repositories.stage_run import StageRun, StageRunRepository
from ...utils.datetime import to_utc_iso_string
from .workflows import make_stage_dto


@dataclass
class DataRequestSelection:
    stage_id: str
    limit: int
    offset: int

    @staticmethod
    def from_dict(data: dict) -> "DataRequestSelection":
        return DataRequestSelection(
            stage_id=data["stage_id"],
            offset=data.get("offset", 0),
            limit=data.get("limit", 10),
        )

    def to_dict(self) -> dict:
        return dict(stage_id=self.stage_id, offset=self.offset, limit=self.limit)


@dataclass
class DataRequestFilter:
    status: Optional[str]
    asignee: Optional[str]
    data: Optional[dict]

    @staticmethod
    def from_dict(data: dict) -> "DataRequestFilter":
        return DataRequestFilter(
            status=data.get("status", None),
            asignee=data.get("asignee", None),
            data=data.get("data", None),
        )

    def to_dict(self) -> dict:
        return dict((k, v) for k, v in self.__dict__.items() if v is not None)


@dataclass
class DataRequest:
    selection: List[DataRequestSelection]
    filter: DataRequestFilter

    @staticmethod
    def from_dict(data: dict) -> "DataRequest":
        return DataRequest(
            selection=[
                DataRequestSelection.from_dict(s) for s in data.get("selection", [])
            ],
            filter=DataRequestFilter.from_dict(data.get("filter", {})),
        )

    def to_dict(self) -> dict:
        return dict(
            selection=[s.to_dict() for s in self.selection],
            filter=self.filter.to_dict(),
        )


@dataclass
class StageCardContentItem:
    key: str
    value: Any
    type: str = "text"

    def to_dict(self) -> dict:
        return dict(
            key=self.key,
            value=self.value,
            type=self.type,
        )


StageCardContent = List[StageCardContentItem]


@dataclass
class StageRunCard:
    id: str
    created_at: datetime
    status: str
    content: StageCardContent

    def to_dict(self) -> dict:
        return dict(
            id=self.id,
            created_at=to_utc_iso_string(self.created_at),
            status=self.status,
            content=[item.to_dict() for item in self.content],
        )


@dataclass
class ColumnStage:
    id: str
    type: str
    path: Optional[str]
    title: str
    can_be_started: bool = False

    @staticmethod
    def create(stage: WorkflowStage) -> "ColumnStage":
        return ColumnStage(
            id=stage.id,
            type=stage.type_name,
            title=stage.title,
            path=stage.path if isinstance(stage, FormStage) else None,
            can_be_started=isinstance(stage, FormStage) and stage.is_initial,
        )

    def to_dict(self) -> dict:
        return dict(
            id=self.id,
            type=self.type,
            title=self.title,
            path=self.path,
            can_be_started=self.can_be_started,
        )


@dataclass
class KanbanColumn:
    selected_stage: ColumnStage
    stage_run_cards: List[StageRunCard]
    stages: List[ColumnStage]
    total_count: int

    def to_dict(self) -> dict:
        return dict(
            selected_stage=self.selected_stage.to_dict(),
            stage_run_cards=[card.to_dict() for card in self.stage_run_cards],
            stages=[stage.to_dict() for stage in self.stages],
            total_count=self.total_count,
        )


@dataclass
class KanbanData:
    columns: List[KanbanColumn]
    next_stage_options: List[ColumnStage]

    def to_dict(self) -> dict:
        return dict(
            columns=[column.to_dict() for column in self.columns],
            next_stage_options=[option.to_dict() for option in self.next_stage_options],
        )


class KanbanController:
    def __init__(
        self,
        stage_run_repository: StageRunRepository,
        project_repository: Type[ProjectRepository],
    ) -> None:
        self.stage_run_repository = stage_run_repository
        self.project_repository = project_repository

    def stage_run_content(self, stage_run: StageRun):
        visualization = self.project_repository.load().visualization
        visualization_defined_data = [
            StageCardContentItem(
                key=v.name,
                value=stage_run.data.get(v.name),
                type=v.type,
            )
            for v in visualization.items
        ]
        visualization_defined_keys = [v.name for v in visualization.items]
        extra_data = [
            StageCardContentItem(key=key, value=stage_run.data.get(key))
            for key in sorted(stage_run.data.keys())
            if key not in visualization_defined_keys
        ]
        return visualization_defined_data + extra_data

    def _get_next_stages(
        self, previous_stage: Optional[str] = None
    ) -> Sequence[WorkflowStage]:
        project = ProjectRepository.load()
        if previous_stage:
            origin = project.get_stage(previous_stage)
            if not origin:
                raise Exception(f"Stage {previous_stage} not found")

            next_stages = []
            for next_stage in origin.workflow_transitions:
                stage = project.get_stage(next_stage.target_id)
                if not stage:
                    raise Exception(f"Stage {next_stage.target_id} not found")
                next_stages.append(stage)

            return next_stages

        else:
            return project.get_initial_stages()

    def _get_column_stages(self, selected_stages_ids: List[str], column_idx: int):
        if column_idx == 0:
            return self._get_next_stages()
        else:
            return self._get_next_stages(selected_stages_ids[column_idx - 1])

    def get_data(self, request: DataRequest) -> KanbanData:
        project = self.project_repository.load()

        selected_stage_ids = [s.stage_id for s in request.selection]
        columns: List[KanbanColumn] = []
        for column_idx, selection in enumerate(request.selection):
            request_filter = GetStageRunByQueryFilter.from_dict(
                data=request.filter.to_dict()
            )
            request_filter.stage = selection.stage_id
            paginated_response = self.stage_run_repository.find_leaves(
                filter=request_filter,
                pagination=Pagination(limit=selection.limit, offset=selection.offset),
            )
            stage_run_cards = [
                StageRunCard(
                    id=stage_run.id,
                    status=stage_run.status,
                    created_at=stage_run.created_at,
                    content=self.stage_run_content(stage_run),
                )
                for stage_run in paginated_response.stage_runs
            ]
            stages = [
                ColumnStage.create(s)
                for s in self._get_column_stages(selected_stage_ids, column_idx)
            ]
            selected_stage_run = ColumnStage.create(
                project.get_stage_raises(selection.stage_id)
            )

            columns.append(
                KanbanColumn(
                    selected_stage=selected_stage_run,
                    stage_run_cards=stage_run_cards,
                    stages=stages,
                    total_count=paginated_response.total_count,
                )
            )

        next_stage_options = [
            ColumnStage.create(s)
            for s in self._get_column_stages(
                selected_stage_ids, len(selected_stage_ids)
            )
        ]

        return KanbanData(columns=columns, next_stage_options=next_stage_options)

    def get_ancestor_logs(
        self, stage_run_id: str
    ) -> List[Tuple[StageRun, List[LogEntry]]]:
        ancestors = self.stage_run_repository.find_ancestors(stage_run_id)
        return [
            (
                ancestor,
                (
                    execution_logs_repository.get(execution_id=ancestor.execution_id)
                    if ancestor.execution_id
                    else []
                ),
            )
            for ancestor in ancestors
        ]


def get_editor_bp():
    project_repository = ProjectRepository

    controller = KanbanController(stage_run_repository, project_repository)
    bp = flask.Blueprint("kanban", __name__)

    @bp.post("/")
    def _get_kanban():
        if flask.request.json is None:
            flask.abort(400)
        req = DataRequest.from_dict(flask.request.json)
        return controller.get_data(req).to_dict()

    @bp.get("/logs/<stage_run_id>")
    def _get_ancestor_logs(stage_run_id: str):
        def stage_from_run(stage_run: StageRun):
            stage = project_repository.load().get_stage_raises(stage_run.stage)
            return make_stage_dto(stage)

        return [
            {
                "stage": stage_from_run(stage_run),
                "stage_run": stage_run.to_dto(),
                "logs": [
                    dict(
                        event=log.event,
                        payload=log.payload,
                        createdAt=to_utc_iso_string(log.created_at),
                    )
                    for log in logs
                    if (
                        isinstance(log, StdioLogEntry)
                        or isinstance(log, UnhandledExceptionLogEntry)
                    )
                    and log.payload["text"].strip() != ""
                ],
            }
            for stage_run, logs in controller.get_ancestor_logs(stage_run_id)
        ]

    return bp


def get_player_bp():
    project_repository = ProjectRepository
    stage_run_repository = stage_run_repository_factory()

    members_repository = members_repository_factory()
    guard = RoleGuardFactory(members_repository)

    controller = KanbanController(stage_run_repository, project_repository)
    bp = flask.Blueprint("kanban", __name__)

    @bp.post("/")
    @guard.requires("workflow_viewer")
    def _get_kanban():
        if flask.request.json is None:
            flask.abort(400)
        req = DataRequest.from_dict(flask.request.json)
        return controller.get_data(req).to_dict()

    @bp.get("/logs/<stage_run_id>")
    @guard.requires("workflow_viewer")
    def _get_ancestor_logs(stage_run_id: str):
        def stage_from_run(stage_run: StageRun):
            stage = project_repository.load().get_stage_raises(stage_run.stage)
            return make_stage_dto(stage)

        return [
            {
                "stage": stage_from_run(stage_run),
                "stage_run": stage_run.to_dto(),
                "logs": [log for log in logs],
            }
            for stage_run, logs in controller.get_ancestor_logs(stage_run_id)
        ]

    return bp



import os
from typing import Dict
from unittest import TestCase
from .env_masker import EnvMasker


def set_env(env: Dict[str, str]) -> None:
    for k, v in env.items():
        os.environ[k] = v


class TestImportAsNew(TestCase):
    def tearDown(self) -> None:
        os.environ.clear()

    def test_values(self):
        set_env({"ENV1": "brown", "ENV2": "FOX", "ENV3": "1892"})
        masker = EnvMasker(min_value_size=2)
        masked = masker.mask("the quick brown fox jumps over the lazy dog - 1892")
        self.assertEqual(
            masked, "the quick <MASKED> <MASKED> jumps over the lazy dog - <MASKED>"
        )

    def test_small_values(self):
        set_env({"ENV1": "fox", "ENV2": "LAZY", "ENV3": "18", "ENV4": "jump"})
        masker = EnvMasker(min_value_size=4)
        masked = masker.mask("the quick brown fox jumps over the lazy dog - 1892")
        self.assertEqual(
            masked, "the quick brown fox <MASKED>s over the <MASKED> dog - 1892"
        )

    def test_multiline(self):
        set_env(
            {
                "ENV1": """---- BEGIN SSH2 PUBLIC KEY ----"""
                + """AAAAB3NzaC1yc2EAAAABJQAAAQB/nAmOjTmezNUDKYvEeIRf2YnwM9/uUG1d0BYs"""
                + """c8/tRtx+RGi7N2lUbp728MXGwdnL9od4cItzky/zVdLZE2cycOa18xBK9cOWmcKS"""
                + """0A8FYBxEQWJ/q9YVUgZbFKfYGaGQxsER+A0w/fX8ALuk78ktP31K69LcQgxIsl7r"""
                + """NzxsoOQKJ/CIxOGMMxczYTiEoLvQhapFQMs3FL96didKr/QbrfB1WT6s3838SEaX"""
                + """fgZvLef1YB2xmfhbT9OXFE3FXvh2UPBfN+ffE7iiayQf/2XR+8j4N4bW30DiPtOQ"""
                + """---- END SSH2 PUBLIC KEY ----"""
            }
        )
        masker = EnvMasker(min_value_size=4)
        masked = masker.mask(f"credentials: {os.environ['ENV1']}")
        self.assertEqual(masked, "credentials: <MASKED>")



import json
from typing import Optional
from ....utils.json import to_json_serializable
from ....execution.execution import UnsetStageRun, get_current_execution_throwable


def is_json_serializable(value):
    try:
        json.dumps(value)
        return True
    except TypeError:
        return False


def get_data(key: Optional[str] = None):
    execution = get_current_execution_throwable()
    stage_run = execution.stage_run
    if not stage_run:
        raise UnsetStageRun()
    return stage_run.data.get(key) if key else stage_run.data


def set_data(key, value):
    value = to_json_serializable(value)
    if not is_json_serializable(value):
        raise TypeError(f"{key} is not JSON serializable")
    execution = get_current_execution_throwable()
    stage_run = execution.stage_run
    if not stage_run:
        raise UnsetStageRun()
    stage_run.data[key] = value



import json
from io import BytesIO

from ...execution import HookExecution


def _app_json_parse(body: str):
    return json.loads(body)


def _multipart_form_parse(body: str, headers: dict):
    from multipart import MultipartParser, parse_options_header, to_bytes

    _, options = parse_options_header(headers["Content-Type"])
    boundary = options["boundary"].encode("utf-8")
    p = MultipartParser(BytesIO(to_bytes(body)), boundary)  # TODO: direct from fs
    return [{"name": i.name, "value": i.value} for i in p]


def get_raw_request():
    execution = HookExecution.get_current_hook_execution()
    request = execution.context.request
    return request.body, request.query_params, request.headers


def get_request():
    body, query, headers = get_raw_request()
    try:
        if "application/json" in headers["Content-Type"]:
            return _app_json_parse(body), query, headers
        elif "multipart/form-data" in headers["Content-Type"]:
            return _multipart_form_parse(body, headers), query, headers
        else:
            return body, query, headers
    except Exception:
        return body, query, headers


def send_response(body="", status_code=200, headers={}):
    execution = HookExecution.get_current_hook_execution()
    execution.context.response = (body, status_code, headers)


def send_json(data={}, status_code=200, headers={}):
    send_response(
        body=json.dumps(data, allow_nan=False),
        headers={**headers, "content-type": "application/json"},
        status_code=status_code,
    )



from unittest import TestCase
from tests.fixtures import init_dir, clear_dir
from .project import (
    NotificationTrigger,
    ProjectRepository,
    ScriptStage,
    Project,
    FormStage,
    WorkflowTransition,
)


class ProjectTests(TestCase):
    def setUp(self):
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_delete_file_if_remove_file_true(self):
        project = ProjectRepository.load()
        file = self.root / "script.py"
        file.write_text("print('hello, world')")

        script = ScriptStage(
            file="script.py",
            id="test",
            is_initial=True,
            title="test",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.scripts.append(script)
        ProjectRepository.save(project)
        project.delete_stage("test", remove_file=True)
        self.assertFalse(file.exists())

    def test_dont_delete_file_if_remove_file_false(self):
        project = ProjectRepository.load()
        file = self.root / "script.py"
        file.write_text("print('hello, world')")

        script = ScriptStage(
            file="script.py",
            id="test",
            is_initial=True,
            title="test",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )
        project.scripts.append(script)
        ProjectRepository.save(project)
        project.delete_stage("test", remove_file=False)
        self.assertTrue(file.exists())

    def test_is_initial_false(self):
        project = Project.create()

        form1 = FormStage(
            id="form1",
            path="form1",
            title="form1",
            file="form1.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        form2 = FormStage(
            id="form2",
            path="form2",
            title="form2",
            file="form2.py",
            workflow_position=(0.0, 0.0),
            workflow_transitions=[],
            notification_trigger=NotificationTrigger(
                variable_name="val", enabled=False
            ),
        )

        script = ScriptStage(
            file="script.py",
            id="script",
            is_initial=True,
            title="script",
            workflow_position=(0, 0),
            workflow_transitions=[],
        )

        form1.workflow_transitions.append(
            WorkflowTransition(
                id="1",
                target_id=form2.id,
                target_type=form2.type_name,
                type=WorkflowTransition.default_type(form1.type_name),
            )
        )

        form1.workflow_transitions.append(
            WorkflowTransition(
                id="2",
                target_id=script.id,
                target_type=script.type_name,
                type=WorkflowTransition.default_type(form1.type_name),
            )
        )

        project = ProjectRepository.load()

        project.add_stage(form1)
        project.add_stage(form2)
        project.add_stage(script)

        ProjectRepository.save(project)

        read_project = ProjectRepository.load()

        read_form1 = read_project.get_stage(form1.id)
        if read_form1 is None or not isinstance(read_form1, FormStage):
            self.fail("form1 is null")

        self.assertTrue(read_form1.is_initial)

        read_form2 = read_project.get_stage(form2.id)
        if read_form2 is None or not isinstance(read_form2, FormStage):
            self.fail("form2 is null")

        self.assertFalse(read_form2.is_initial)

        read_script = read_project.get_stage(script.id)
        if read_script is None or not isinstance(read_script, ScriptStage):
            self.fail("script is null")

        self.assertFalse(read_script.is_initial)



import flask
from typing import Callable
from functools import wraps
from ...contracts_generated import CommonUserRolesItem
from ...repositories.members import MembersRepository
from ...jwt_auth import UserClaims
from typing import Callable, Optional


def default_auth_decoder(auth_header: str) -> Optional[UserClaims]:
    return UserClaims.from_jwt(auth_header.split(" ")[1])


class RoleGuardFactory:
    repository: MembersRepository
    auth_decoder: Callable[[str], Optional[UserClaims]]

    def __init__(
        self,
        repository: MembersRepository,
        auth_decoder: Callable[[str], Optional[UserClaims]] = default_auth_decoder,
    ) -> None:
        self.repository = repository
        self.auth_decoder = auth_decoder

    def requires(self, role: CommonUserRolesItem):
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                auth = flask.request.headers.get("Authorization")
                if auth is None:
                    flask.abort(401)

                claims = self.auth_decoder(auth)
                if claims is None or claims.email is None:
                    flask.abort(401)

                user_roles = self.repository.get_member_roles(claims.email)
                if len(user_roles) == 0 or role not in user_roles:
                    flask.abort(403)

                return func(*args, **kwargs)

            return wrapper

        return decorator
